#include "cstd.h"
#include "rngs.h"

/* 2,4 staggered grid acoustic scheme, definitions of forward,
   linearized and adjoint linearized maps, dot product tests for all
   components. Based in simplified structs resembling those in
   IWAVE. Dirichlet conditions only.

   WWS 02.12
 */

/* constants */
#define NZ 100           // grid size - z
#define NX 100           // grid size - x
#define DZ 10            // step - z
#define DX 10            // step - x
#define ISZ 2            // source z index
#define ISX 50           // source x index
#define NG 100           // number of recvrs
#define GZ 1             // recvr z index
#define GXMIN 0          // min recvr x index
#define NT 500           // number of time steps
#define DT 2             // step - t
#define MPBG 2250        // background value of bulk mod
#define MVBG 0.001       // background value of buoyancy
#define MPVAR 1000       // variance of bulk mod about bockground
#define MVVAR 0.0005     // variance of buoyancy about background
#define DMPVAR 1.0       // variance of bulk mod, buoyancy perts

#define OHOH 0

// analogue of iwave RDOM - flat data structure
typedef struct duhgrid {
  int nz;
  int nx;
  float dz;
  float dx;
  float * mp;
  int mp_offz;
  int mp_offx;
  int mp_nz;
  int mp_nx;
  float * mvz;
  int mvz_offz;
  int mvz_offx;
  int mvz_nz;
  int mvz_nx;
  float * mvx;
  int mvx_offz;
  int mvx_offx;
  int mvx_nz;
  int mvx_nx;
  float * p;
  int p_offz;
  int p_offx;
  int p_nz;
  int p_nx;
  float * vz;
  int vz_offz;
  int vz_offx;
  int vz_nz;
  int vz_nx;
  float * vx;
  int vx_offz;
  int vx_offx;
  int vx_nz;
  int vx_nx;
} duhgrid;

// source coords
typedef struct duhsrcs {
  int isz;
  int isx;
} duhsrcs;

// analogue of iwave tracegeom
typedef struct duhrecs { 
  int ng;
  int nt;
  float dt;
  int * igz;
  int * igx;
  float * data;
} duhrecs;

// history structure - not self-descriptive
typedef struct hstr {
  float * p;
  float * vz;
  float * vx;
} hstr;

void field_init(float ** f, int n, float bg, float var) {
  int i;
  *f = (float *)malloc(n*sizeof(float));
  for (i=0;i<n;i++) 
    (*f)[i]=bg+var*((float)Random()-0.5f);
}

void duhgrid_init(duhgrid * g) {

  g->mp_offz = 1;
  g->mp_offx = 1;
  g->mp_nz   = NZ;
  g->mp_nx   = NX;
  g->mvz_offz = 0;
  g->mvz_offx = 1;
  g->mvz_nz   = NZ+1;
  g->mvz_nx   = NX;
  g->mvx_offz = 1;
  g->mvx_offx = 0;
  g->mvx_nz   = NZ;
  g->mvx_nx   = NX+1;
  g->p_offz  = -1;
  g->p_offx  = -1;
  g->p_nz    = NZ+4;
  g->p_nx    = NX+4;
  g->vz_offz = -1;
  g->vz_offx = 1;
  g->vz_nz   = NZ+3;
  g->vz_nx   = NX;
  g->vx_offz = 1;
  g->vx_offx = -1;
  g->vx_nz   = NZ;
  g->vx_nx   = NX+3;

  g->nz=NZ+2;
  g->nx=NX+2;
  g->dz=DZ;
  g->dx=DX;
  field_init(&(g->mp),(g->mp_nz)*(g->mp_nx),MPBG,MPVAR);
  field_init(&(g->mvz),(g->mvz_nz)*(g->mvz_nx),MVBG,MVVAR);
  field_init(&(g->mvx),(g->mvx_nz)*(g->mvx_nx),MVBG,MVVAR);
  field_init(&(g->p),(g->p_nz)*(g->p_nx),0.0f,0.0f);
  field_init(&(g->vz),(g->vz_nz)*(g->vz_nx),0.0f,0.0f);
  field_init(&(g->vx),(g->vx_nz)*(g->vx_nx),0.0f,0.0f);

}

void static_field_copy(duhgrid * tgt, duhgrid * src) {
  int i;
  for (i=0;i<src->mp_nz*src->mp_nx;i++) tgt->mp[i]=src->mp[i];
  for (i=0;i<src->mvz_nz*src->mvz_nx;i++) tgt->mvz[i]=src->mvz[i];
  for (i=0;i<src->mvx_nz*src->mvx_nx;i++) tgt->mvx[i]=src->mvx[i];
}

void static_field_rand(duhgrid * tgt, float var) {
  int i;
  for (i=0;i<tgt->mp_nz*tgt->mp_nx;i++) tgt->mp[i]=Random();
  for (i=0;i<tgt->mvz_nz*tgt->mvz_nx;i++) tgt->mvz[i]=Random();
  for (i=0;i<tgt->mvx_nz*tgt->mvx_nx;i++) tgt->mvx[i]=Random();
}

void static_field_zero(duhgrid * tgt) {
  int i;
  for (i=0;i<tgt->mp_nz*tgt->mp_nx;i++) tgt->mp[i]=0.0f;
  for (i=0;i<tgt->mvz_nz*tgt->mvz_nx;i++) tgt->mvz[i]=0.0f;
  for (i=0;i<tgt->mvx_nz*tgt->mvx_nx;i++) tgt->mvx[i]=0.0f;
}

float static_field_ip(duhgrid * tgt, duhgrid * src) {
  int i;
  float dot = 0.0;
  for (i=0;i<src->mp_nz*src->mp_nx;i++) dot += tgt->mp[i]*src->mp[i];
  for (i=0;i<src->mvz_nz*src->mvz_nx;i++) dot += tgt->mvz[i]*src->mvz[i];
  for (i=0;i<src->mvx_nz*src->mvx_nx;i++) dot += tgt->mvx[i]*src->mvx[i];
  dot *= tgt->dz*tgt->dx;
  return dot;
}

void dynamic_field_copy(duhgrid * tgt, duhgrid * src) {
  int i;
  for (i=0;i<src->p_nz*src->p_nx;i++) tgt->p[i]=src->p[i];
  for (i=0;i<src->vz_nz*src->vz_nx;i++) tgt->vz[i]=src->vz[i];
  for (i=0;i<src->vx_nz*src->vx_nx;i++) tgt->vx[i]=src->vx[i];
}

void dynamic_field_zero(duhgrid * tgt) {
  int i;
  for (i=0;i<tgt->p_nz*tgt->p_nx;i++) tgt->p[i]=0.0f;
  for (i=0;i<tgt->vz_nz*tgt->vz_nx;i++) tgt->vz[i]=0.0f;
  for (i=0;i<tgt->vx_nz*tgt->vx_nx;i++) tgt->vx[i]=0.0f;
}

void dynamic_field_rand(duhgrid * tgt, float var) {
  int i;
  int iz,ix;

  for (i=0;i<tgt->p_nz*tgt->p_nx;i++) tgt->p[i]=Random();
  for (i=0;i<tgt->vz_nz*tgt->vz_nx;i++) tgt->vz[i]=Random();
  for (i=0;i<tgt->vx_nz*tgt->vx_nx;i++) tgt->vx[i]=Random();

  /* odd pressure, even velocity mirroring */

  for (ix=0; ix<tgt->p_nx; ix++) {
    tgt->p[ix*tgt->p_nz]=-tgt->p[2+ix*tgt->p_nz];
    tgt->p[1+ix*tgt->p_nz]= 0.0f;
    tgt->p[tgt->p_nz-1+ix*tgt->p_nz]=-tgt->p[tgt->p_nz-3+ix*tgt->p_nz];
    tgt->p[tgt->p_nz-2+ix*tgt->p_nz]= 0.0f;
  }
  for (ix=0; ix<tgt->vz_nx; ix++) {
    tgt->vz[ix*tgt->vz_nz]=tgt->vz[1+ix*tgt->vz_nz];
    tgt->vz[tgt->vz_nz-1+ix*tgt->vz_nz]=tgt->vz[tgt->vz_nz-2+ix*tgt->vz_nz];
  }
  for (iz=0; iz<tgt->p_nz; iz++) {
    tgt->p[iz]=-tgt->p[iz+2*tgt->p_nz]; 
    tgt->p[iz+tgt->p_nz]= 0.0f;
    tgt->p[iz+(tgt->p_nx-1)*tgt->p_nz]=-tgt->p[iz+(tgt->p_nx-3)*tgt->p_nz];
    tgt->p[iz+(tgt->p_nx-2)*tgt->p_nz]= 0.0f;
  }
  for (iz=0; iz<tgt->vx_nz; iz++)  {
    tgt->vx[iz]=tgt->vx[iz+tgt->vx_nz];
    tgt->vx[iz+(tgt->vx_nx-1)*tgt->vx_nz]=tgt->vx[iz+(tgt->vx_nx-2)*tgt->vx_nz];
  }

}

float dynamic_field_ip(duhgrid * tgt, duhgrid * src) {

  int i;
  float dot = 0.0;
  
  for (i=0;i<src->p_nz*src->p_nx;i++) dot += tgt->p[i]*src->p[i];
  for (i=0;i<src->vz_nz*src->vz_nx;i++) dot += tgt->vz[i]*src->vz[i];
  for (i=0;i<src->vx_nz*src->vx_nx;i++) dot += tgt->vx[i]*src->vx[i];
  
  dot *= tgt->dz*tgt->dx;
  return dot;
}

void dynamic_field_write(duhgrid * g) {
  int iz,ix;

  printf("-------------------\n");

  printf("P: nz=%d nx=%d offz=%d offx=%d\n",g->p_nz,g->p_nx,g->p_offz,g->p_offx);
  for (ix=0;ix<g->p_nx;ix++) {
    for (iz=0;iz<g->p_nz;iz++) {
      printf("%10.4e ",g->p[iz+ix*g->p_nz]);
    }
    printf("\n");
  }

  printf("VZ: nz=%d nx=%d offz=%d offx=%d\n",g->vz_nz,g->vz_nx,g->vz_offz,g->vz_offx);
  for (ix=0;ix<g->vz_nx;ix++) {
    for (iz=0;iz<g->vz_nz;iz++) {
      printf("%10.4e ",g->vz[iz+ix*g->vz_nz]);
    }
    printf("\n");
  }

  printf("VX: nz=%d nx=%d offz=%d offx=%d\n",g->vx_nz,g->vx_nx,g->vx_offz,g->vx_offx);
  for (ix=0;ix<g->vx_nx;ix++) {
    for (iz=0;iz<g->vx_nz;iz++) {
      printf("%10.4e ",g->vx[iz+ix*g->vx_nz]);
    }
    printf("\n");
  }

  printf("-------------------\n");
}

void duhgrid_destroy(duhgrid * g) {
  free(g->mp);
  free(g->mvz);
  free(g->mvx);
  free(g->p);
  free(g->vz);
  free(g->vx);
}

void hstr_init(hstr * h, duhgrid * m, duhrecs * t) {
  int plen = m->p_nz*m->p_nx*t->nt;
  int vzlen = m->vz_nz*m->vz_nx*t->nt;
  int vxlen = m->vx_nz*m->vx_nx*t->nt;
  h->p = (float *)malloc(plen*sizeof(float));
  h->vz = (float *)malloc(vzlen*sizeof(float));
  h->vx = (float *)malloc(vxlen*sizeof(float));
}

void hstr_destroy(hstr * h) {
  free(h->p);
  free(h->vz);
  free(h->vx);
}

void duhsrcs_init(duhsrcs * s) {

  s->isz = ISZ;
  s->isx = ISX;

}

void duhrecs_init(duhrecs * r) {
  
  int ig;

  r->ng = NG;
  r->nt = NT;
  r->dt = DT;

  field_init(&(r->data),NT*NG,0.0,0.0);

  r->igz = (int *)malloc(NG*sizeof(int));
  r->igx = (int *)malloc(NG*sizeof(int));
  for (ig=0;ig<r->ng;ig++) {
    (r->igz)[ig]=GZ;
    (r->igx)[ig]=GXMIN+ig;
  }
}

void duhrecs_zero(duhrecs *r) {
  int i;
  for (i=0;i<r->nt*r->ng;i++) r->data[i]=0.0f;
}

void duhrecs_rand(duhrecs *r, float var) {
  int i;
  for (i=0;i<r->nt*r->ng;i++) r->data[i]=var*(Random()-0.5f);
}

float duhrecs_ip(duhrecs * r1, duhrecs * r2) {
  int i;
  float dot = 0.0f;
  for (i=0;i<r1->nt*r1->ng;i++) 
    dot += r1->data[i]*r2->data[i];
  dot *= r1->dt;
  return dot;
}

void duhrecs_write(duhrecs * r, const char * name) {

  char * hfile;
  char * dfile;
  FILE * fph;
  FILE * fpd;
  int l = strlen(name)+20;
  
  hfile = (char *)malloc(l*sizeof(char));
  dfile = (char *)malloc(l*sizeof(char));
  
  strcpy(hfile,name);
  strcat(hfile,".rsf");
  strcpy(dfile,name);
  strcat(dfile,".rsf@");

  fph=fopen(hfile,"w");
  fprintf(fph,"n1=%d\n",r->nt);
  fprintf(fph,"d1=%10.4e\n",r->dt);
  fprintf(fph,"o1=0.0\n");
  fprintf(fph,"n2=%d\n",r->ng);
  fprintf(fph,"d2=1.0\no2=0.0\n");
  fprintf(fph,"data_format=native_float\n");
  fprintf(fph,"data_type=pressure\n");
  fprintf(fph,"in=%s\n",dfile);
  fclose(fph);

  fpd=fopen(dfile,"w");
  fwrite(r->data,sizeof(float),r->nt*r->ng,fpd);
  fclose(fpd);

}
  
void duhrecs_destroy(duhrecs * r) {
  free(r->data);
  free(r->igz);
  free(r->igx);
}

void duhsrcs_to_duhgrid(duhsrcs * s, duhgrid * g, int it) {
  if (it==0) g->p[s->isz-g->p_offz+g->p_nz*(s->isx-g->p_offx)] += 1.0f/DT;
}

void duhgrid_to_duhrecs(duhrecs * r, duhgrid * g, int it) {
  int ig;
  for (ig=0;ig<r->ng;ig++) 
    r->data[it+ig*r->nt] += g->p[r->igz[ig]-g->p_offz+g->p_nz*(r->igx[ig]-g->p_offx)];
}

void duhrecs_to_duhgrid(duhrecs * r, duhgrid * g, int it) {
  int ig;
  float fac = r->dt/(g->dx*g->dz);
  for (ig=0;ig<r->ng;ig++) 
    g->p[r->igz[ig]-g->p_offz+g->p_nz*(r->igx[ig]-g->p_offx)] += fac*((r->data)[it+ig*r->nt]);
}

void duhgrid_to_hist(hstr * h, duhgrid * m, int it) {
  int i;
  int start;
  start=it*m->p_nz*m->p_nx;
  for (i=0;i<m->p_nz*m->p_nx;i++) h->p[i+start]=m->p[i];
  start=it*m->vz_nz*m->vz_nx;
  for (i=0;i<m->vz_nz*m->vz_nx;i++) h->vz[i+start]=m->vz[i];
  start=it*m->vx_nz*m->vx_nx;
  for (i=0;i<m->vx_nz*m->vx_nx;i++) h->vx[i+start]=m->vx[i];
}

void hist_to_duhgrid(hstr * h, duhgrid * m, int it) {
  int i;
  int start;
  start=it*m->p_nz*m->p_nx;
  for (i=0;i<m->p_nz*m->p_nx;i++) m->p[i]=h->p[i+start];
  start=it*m->vz_nz*m->vz_nx;
  for (i=0;i<m->vz_nz*m->vz_nx;i++) m->vz[i]=h->vz[i+start];
  start=it*m->vx_nz*m->vx_nx;
  for (i=0;i<m->vx_nz*m->vx_nx;i++) m->vx[i]=h->vx[i+start];
}

void fts_p(duhgrid * g, float dt) {

  int ix, iz;

  float c1  = -27.0/24.0;
  float c2  = 1.0/24.0;
  float lamz = dt/g->dz;
  float lamx = dt/g->dx;
  float *p = g->p;
  float *mp = g->mp;
  float *vz = g->vz;
  float *vx = g->vx;

  int ip;
  int imp;
  int ivz;
  int ivx;
  int divz;
  int divx;

  for (ix=1;ix<NX;ix++) {
    for (iz=1;iz<NZ;iz++) {

      ip  = iz-g->p_offz + (ix-g->p_offx)*g->p_nz;
      imp = iz-g->mp_offz + (ix-g->mp_offx)*g->mp_nz; 
      ivz = iz-g->vz_offz + (ix-g->vz_offx)*g->vz_nz;
      ivx = iz-g->vx_offz + (ix-g->vx_offx)*g->vx_nz;
      divz = 1;
      divx = g->vx_nz;
      
      p[ip] += mp[imp]*
	(lamz*(c1*(vz[ivz]-vz[ivz-divz])+c2*(vz[ivz+divz]-vz[ivz-2*divz])) +
	 lamx*(c1*(vx[ivx]-vx[ivx-divx])+c2*(vx[ivx+divx]-vx[ivx-2*divx])));
	
    }
  }

  for (ix=0; ix<g->p_nx; ix++) {
    g->p[ix*g->p_nz]=-g->p[2+ix*g->p_nz];
    g->p[1+ix*g->p_nz]= 0.0f;
    g->p[g->p_nz-1+ix*g->p_nz]=-g->p[g->p_nz-3+ix*g->p_nz];
    g->p[g->p_nz-2+ix*g->p_nz]= 0.0f;
  }
  for (iz=0; iz<g->p_nz; iz++) {
    g->p[iz]=-g->p[iz+2*g->p_nz]; 
    g->p[iz+g->p_nz]= 0.0f;
    g->p[iz+(g->p_nx-1)*g->p_nz]=-g->p[iz+(g->p_nx-3)*g->p_nz];
    g->p[iz+(g->p_nx-2)*g->p_nz]= 0.0f;
  }

}

void fts_v(duhgrid * g, float dt) {

  int ix, iz;

  float c1  = -27.0/24.0;
  float c2  = 1.0/24.0;
  float lamz = dt/g->dz;
  float lamx = dt/g->dx;
  float *p = g->p;
  float *mvz = g->mvz;
  float *mvx = g->mvx;
  float *vz = g->vz;
  float *vx = g->vx;

  int ip;
  int ivz;
  int ivx;
  int dip;
  int imvz;
  int imvx;

  for (ix=1; ix<NX; ix++) {
    for (iz=0; iz<NZ; iz++) {

      ip  = iz-g->p_offz + (ix-g->p_offx)*g->p_nz;
      dip = 1;
      imvz = iz-g->mvz_offz + (ix-g->mvz_offx)*g->mvz_nz; 
      ivz = iz-g->vz_offz + (ix-g->vz_offx)*g->vz_nz;

      vz[ivz] += mvz[imvz]*lamz*(c1*(p[ip+dip]-p[ip])+c2*(p[ip+2*dip]-p[ip-dip]));
    }
  }
  for (ix=0; ix<g->vz_nx; ix++) {
    g->vz[ix*g->vz_nz]=g->vz[1+ix*g->vz_nz];
    g->vz[g->vz_nz-1+ix*g->vz_nz]=g->vz[g->vz_nz-2+ix*g->vz_nz];
  }

  for (ix=0; ix<NX; ix++) {
    for (iz=1; iz<NZ; iz++) {
      
      ip  = iz-g->p_offz + (ix-g->p_offx)*g->p_nz;
      dip = g->p_nz;
      imvx = iz-g->mvx_offz + (ix-g->mvx_offx)*g->mvx_nz; 
      ivx = iz-g->vx_offz + (ix-g->vx_offx)*g->vx_nz;
      
      vx[ivx] += mvx[imvx]*lamx*(c1*(p[ip+dip]-p[ip])+c2*(p[ip+2*dip]-p[ip-dip]));

    }
  }
  for (iz=0; iz<g->vx_nz; iz++)  {
    g->vx[iz]=g->vx[iz+g->vx_nz];
    g->vx[iz+(g->vx_nx-1)*g->vx_nz]=g->vx[iz+(g->vx_nx-2)*g->vx_nz];
  }

}

void ftsm_p(duhgrid * u, duhgrid * r, float dt) {

  int ix, iz;

  float c1  = -27.0/24.0;
  float c2  = 1.0/24.0;
  float lamz = dt/r->dz;
  float lamx = dt/r->dx;

  int ip;
  int imp;
  int ivz;
  int ivx;
  int divz;
  int divx;

  for (ix=1;ix<NX;ix++) {
    for (iz=1;iz<NZ;iz++) {

      ip  = iz-r->p_offz + (ix-r->p_offx)*r->p_nz;
      imp = iz-r->mp_offz + (ix-r->mp_offx)*r->mp_nz; 
      ivz = iz-r->vz_offz + (ix-r->vz_offx)*r->vz_nz;
      ivx = iz-r->vx_offz + (ix-r->vx_offx)*r->vx_nz;
      divz = 1;
      divx = r->vx_nz;
      
      u->p[ip] += 
	// step 1: state derivative 
	r->mp[imp]*
	(lamz*(c1*(u->vz[ivz]-u->vz[ivz-divz])+c2*(u->vz[ivz+divz]-u->vz[ivz-2*divz])) +
	 lamx*(c1*(u->vx[ivx]-u->vx[ivx-divx])+c2*(u->vx[ivx+divx]-u->vx[ivx-2*divx])))
	// step 2: control derivative
	+ 
	u->mp[imp]*
	(lamz*(c1*(r->vz[ivz]-r->vz[ivz-divz])+c2*(r->vz[ivz+divz]-r->vz[ivz-2*divz])) +
	 lamx*(c1*(r->vx[ivx]-r->vx[ivx-divx])+c2*(r->vx[ivx+divx]-r->vx[ivx-2*divx])));
	
    }
  }

  // reflection bcs
  for (ix=0; ix<r->p_nx; ix++) {
    u->p[ix*r->p_nz]=-u->p[2+ix*r->p_nz];
    u->p[1+ix*r->p_nz]= 0.0f;
    u->p[r->p_nz-1+ix*r->p_nz]=-u->p[r->p_nz-3+ix*r->p_nz];
    u->p[r->p_nz-2+ix*r->p_nz]= 0.0f;
  }

  for (iz=0; iz<r->p_nz; iz++) {
    u->p[iz]=-u->p[iz+2*r->p_nz]; 
    u->p[iz+r->p_nz]= 0.0f;
    u->p[iz+(r->p_nx-1)*r->p_nz]=-u->p[iz+(r->p_nx-3)*r->p_nz];
    u->p[iz+(r->p_nx-2)*r->p_nz]= 0.0f;
  }
}

void ftsm_v(duhgrid * u, duhgrid * r, float dt) {

  int ix, iz;

  float c1  = -27.0/24.0;
  float c2  = 1.0/24.0;
  float lamz = dt/r->dz;
  float lamx = dt/r->dx;

  int ip;
  int dip;
  int ivz;
  int ivx;
  int imvz;
  int imvx;

  for (ix=1; ix<NX; ix++) {
    for (iz=0; iz<NZ; iz++) {

      ip  = iz-r->p_offz + (ix-r->p_offx)*r->p_nz;
      dip = 1;
      imvz = iz-r->mvz_offz + (ix-r->mvz_offx)*r->mvz_nz; 
      ivz = iz-r->vz_offz + (ix-r->vz_offx)*r->vz_nz;

      u->vz[ivz] += 
	// step 1: state derivative
	r->mvz[imvz]*
	lamz*(c1*(u->p[ip+dip]-u->p[ip])+c2*(u->p[ip+2*dip]-u->p[ip-dip]))
	// step 2: control derivative
	+
	u->mvz[imvz]*
	lamz*(c1*(r->p[ip+dip]-r->p[ip])+c2*(r->p[ip+2*dip]-r->p[ip-dip]));
    }
  }

  for (ix=0;ix<r->vz_nx;ix++) {
    u->vz[ix*r->vz_nz]=u->vz[1+ix*r->vz_nz];
    u->vz[r->vz_nz-1+ix*r->vz_nz]=u->vz[r->vz_nz-2+ix*r->vz_nz];
  }

  for (ix=0; ix<NX; ix++) {
    for (iz=1; iz<NZ; iz++) {
      
      ip  = iz-r->p_offz + (ix-r->p_offx)*r->p_nz;
      dip = r->p_nz;
      imvx = iz-r->mvx_offz + (ix-r->mvx_offx)*r->mvx_nz; 
      ivx = iz-r->vx_offz + (ix-r->vx_offx)*r->vx_nz;
      
      u->vx[ivx] += 
	// step 1: state derivative
	r->mvx[imvx]*
	lamx*(c1*(u->p[ip+dip]-u->p[ip])+c2*(u->p[ip+2*dip]-u->p[ip-dip]))
	// step 2: control derivative
	+
	u->mvx[imvx]*
	lamx*(c1*(r->p[ip+dip]-r->p[ip])+c2*(r->p[ip+2*dip]-r->p[ip-dip]));

    }
  }

  for (iz=0; iz<r->vx_nz; iz++)  {
    u->vx[iz]=u->vx[iz+r->vx_nz];
    u->vx[iz+(r->vx_nx-1)*r->vx_nz]=u->vx[iz+(r->vx_nx-2)*r->vx_nz];
  }

}

void atsm_v(duhgrid * u, duhgrid * r, float dt) {

  int ix, iz;

  float c1  = -27.0/24.0;
  float c2  = 1.0/24.0;
  float lamz = dt/r->dz;
  float lamx = dt/r->dx;

  int ip;
  int ivz;
  int ivx;
  int dip;
  int imvz;
  int imvx;

  float fac1;
  float fac2;

  // adjoint reflection bcs, vx
  for (iz=0; iz<r->vx_nz; iz++)  {
    u->vx[iz+r->vx_nz]=    u->vx[iz]+u->vx[iz+r->vx_nz];
    u->vx[iz]=0.0f;
    u->vx[iz+(r->vx_nx-2)*r->vx_nz]=    u->vx[iz+(r->vx_nx-1)*r->vx_nz]+u->vx[iz+(r->vx_nx-2)*r->vx_nz];
    u->vx[iz+(r->vx_nx-1)*r->vx_nz]=0.0f;
  }

  // adjoint deriv, p->vx
  for (ix=0; ix<NX; ix++) {
    for (iz=1; iz<NZ; iz++) {
      
      ip  = iz-r->p_offz + (ix-r->p_offx)*r->p_nz;
      dip = r->p_nz;
      imvx = iz-r->mvx_offz + (ix-r->mvx_offx)*r->mvx_nz; 
      ivx = iz-r->vx_offz + (ix-r->vx_offx)*r->vx_nz;
      
      // step 1: adjoint state deriv, p->vx
      //      vx[ivx] += mvx[imvx]*lamx*(c1*(p[ip+dip]-p[ip])+c2*(p[ip+2*dip]-p[ip-dip]));
      fac1=r->mvx[imvx]*lamx;
      fac2=fac1*c2;
      fac1=fac1*c1;
      u->p[ip+dip]    += fac1*u->vx[ivx];
      u->p[ip]        += -fac1*u->vx[ivx];
      u->p[ip+2*dip]  += fac2*u->vx[ivx];
      u->p[ip-dip]    += -fac2*u->vx[ivx];
      // step 2: adjoint control deriv, p->vx
      u->mvx[imvx] += 
	u->vx[ivx]*
	lamx*(c1*(r->p[ip+dip]-r->p[ip])+c2*(r->p[ip+2*dip]-r->p[ip-dip]));
    }
  }

  // adjoint reflection bcs, vz
  for (ix=0; ix<r->vz_nx; ix++) {
    u->vz[1+ix*r->vz_nz] = u->vz[ix*r->vz_nz]+u->vz[1+ix*r->vz_nz];
    u->vz[ix*r->vz_nz]=0.0f;
    u->vz[r->vz_nz-2+ix*r->vz_nz]= u->vz[r->vz_nz-1+ix*r->vz_nz]+u->vz[r->vz_nz-2+ix*r->vz_nz];
    u->vz[r->vz_nz-1+ix*r->vz_nz]=0.0f;
  }

  // adjoint deriv, p->vz
  for (ix=1; ix<NX; ix++) {
    for (iz=0; iz<NZ; iz++) {

      ip  = iz-r->p_offz + (ix-r->p_offx)*r->p_nz;
      dip = 1;
      imvz = iz-r->mvz_offz + (ix-r->mvz_offx)*r->mvz_nz; 
      ivz = iz-r->vz_offz + (ix-r->vz_offx)*r->vz_nz;

      // step 1: adjoint state deriv, p->vz

      //      vz[ivz] += mvz[imvz]*lamz*(c1*(p[ip+dip]-p[ip])+c2*(p[ip+2*dip]-p[ip-dip]));
      fac1=r->mvz[imvz]*lamz;
      fac2=fac1*c2;
      fac1=fac1*c1;
      u->p[ip+dip]    += fac1*u->vz[ivz];
      u->p[ip]        -= fac1*u->vz[ivz];
      u->p[ip+2*dip]  += fac2*u->vz[ivz];
      u->p[ip-dip]    -= fac2*u->vz[ivz];

      // step 2: adjoint control derivative, p->vz
      u->mvz[imvz] += 
	u->vz[ivz]*
	lamz*(c1*(r->p[ip+dip]-r->p[ip])+c2*(r->p[ip+2*dip]-r->p[ip-dip]));
    }
  }
}

void atsm_p(duhgrid * u, duhgrid * r, float dt) {

  int ix, iz;

  float c1  = -27.0/24.0;
  float c2  = 1.0/24.0;
  float lamz = dt/r->dz;
  float lamx = dt/r->dx;

  int ip;
  int imp;
  int ivz;
  int ivx;
  int divz;
  int divx;

  float fac1;
  float fac2;

  // adjoint reflection bcs, p
  for (iz=0; iz<r->p_nz; iz++) {
    u->p[iz+2*r->p_nz] =     -u->p[iz]+u->p[iz+2*r->p_nz]; 
    u->p[iz]=0.0f;
    u->p[iz+r->p_nz]= 0.0f;
    u->p[iz+(r->p_nx-3)*r->p_nz] =     -u->p[iz+(r->p_nx-1)*r->p_nz]+u->p[iz+(r->p_nx-3)*r->p_nz];
    u->p[iz+(r->p_nx-1)*r->p_nz]=0.0f;
    u->p[iz+(r->p_nx-2)*r->p_nz]= 0.0f;
  }
  for (ix=0; ix<r->p_nx; ix++) {
    u->p[2+ix*r->p_nz]= -u->p[ix*r->p_nz] + u->p[2+ix*r->p_nz];
    u->p[ix*r->p_nz]  = 0.0f;
    u->p[1+ix*r->p_nz]= 0.0f;
    u->p[r->p_nz-3+ix*r->p_nz] = -u->p[r->p_nz-1+ix*r->p_nz]+u->p[r->p_nz-3+ix*r->p_nz];
    u->p[r->p_nz-1+ix*r->p_nz] = 0.0f;
    u->p[r->p_nz-2+ix*r->p_nz]= 0.0f;
  }

  // adjoint deriv, vz,vx->p
  for (ix=1;ix<NX;ix++) {
    for (iz=1;iz<NZ;iz++) {

      ip  = iz-r->p_offz + (ix-r->p_offx)*r->p_nz;
      imp = iz-r->mp_offz + (ix-r->mp_offx)*r->mp_nz; 
      ivz = iz-r->vz_offz + (ix-r->vz_offx)*r->vz_nz;
      ivx = iz-r->vx_offz + (ix-r->vx_offx)*r->vx_nz;
      divz = 1;
      divx = r->vx_nz;
      
      // step 1: adjoint state deriv, vz,vx->p
      //p[ip] += mp[imp]*
      //	(lamz*(c1*(vz[ivz]-vz[ivz-divz])+c2*(vz[ivz+divz]-vz[ivz-2*divz])) +
      //	 lamx*(c1*(vx[ivx]-vx[ivx-divx])+c2*(vx[ivx+divx]-vx[ivx-2*divx])));
      fac1=r->mp[imp]*lamz;
      fac2=fac1*c2;
      fac1=fac1*c1;
      u->vz[ivz]        += fac1*u->p[ip];
      u->vz[ivz-divz]   -= fac1*u->p[ip];
      u->vz[ivz+divz]   += fac2*u->p[ip];
      u->vz[ivz-2*divz] -= fac2*u->p[ip];
      fac1=r->mp[imp]*lamx;
      fac2=fac1*c2;
      fac1=fac1*c1;
      u->vx[ivx]        += fac1*u->p[ip];
      u->vx[ivx-divx]   -= fac1*u->p[ip];
      u->vx[ivx+divx]   += fac2*u->p[ip];
      u->vx[ivx-2*divx] -= fac2*u->p[ip];    

      // step 2: adjoint control deriv, vz,vx->p
      u->mp[imp] += 
	u->p[ip]*
	(lamz*(c1*(r->vz[ivz]-r->vz[ivz-divz])+c2*(r->vz[ivz+divz]-r->vz[ivz-2*divz])) +
	 lamx*(c1*(r->vx[ivx]-r->vx[ivx-divx])+c2*(r->vx[ivx+divx]-r->vx[ivx-2*divx])));

    }
  }

}

int main(int argc, char ** argv) {

  int it;
  float aip,bip,aa;
  duhgrid mdl;
  duhrecs traces;
  duhsrcs source;
  hstr hist;

  duhgrid dmdl;
  duhgrid dmin;
  duhgrid dmout;
  duhrecs dtraces;
  duhrecs dtraces1;

  PlantSeeds(getpid());
  
  /* forward map */

  duhgrid_init(&mdl);
  duhrecs_init(&traces);
  duhsrcs_init(&source);
  hstr_init(&hist,&mdl,&traces);
  
  for (it=0;it<traces.nt;it++) {
    // update p;
    fts_p(&mdl,traces.dt);
    // copy entire state to history - p updated, v not yet
    duhgrid_to_hist(&hist,&mdl,it);
    // add spource to p
    duhsrcs_to_duhgrid(&source,&mdl,it);
    // update v
    fts_v(&mdl,traces.dt);
    // record trace
    duhgrid_to_duhrecs(&traces,&mdl,it+1);
  }

  duhrecs_write(&traces,"fwd");

  /* linearized forward map */
  
  duhgrid_init(&dmdl);
  duhgrid_init(&dmin);
  duhgrid_init(&dmout);
  duhrecs_init(&dtraces);
  duhrecs_init(&dtraces1);

  static_field_copy(&dmdl,&mdl);
  

  printf("\n\n\n*** ADJEXPL: standalone acoustic s/g adjoint exploration ***\n");
  /* adjoint test - state deriv, p */
  hist_to_duhgrid(&hist,&mdl,traces.nt/2);

  dynamic_field_rand(&dmdl,DMPVAR);
  // save for later!
  dynamic_field_copy(&dmout,&dmdl);
  static_field_zero(&dmdl);
  
  dynamic_field_rand(&dmin,DMPVAR);
  static_field_zero(&dmin);

  ftsm_p(&dmdl,&mdl,traces.dt);
  aip=dynamic_field_ip(&dmin,&dmdl);
  printf("\nP ONLY:\n");
  printf("<u1,H[c]u0>  =%20.14e\n",aip);
  atsm_p(&dmin,&mdl,traces.dt);
  bip=dynamic_field_ip(&dmin,&dmout);
  printf("<H[c]^Tu1,u0>=%20.14e\n",bip);
  aa=sqrt(dynamic_field_ip(&dmin,&dmin)*dynamic_field_ip(&dmout,&dmout));
  printf("relerr =%20.14e\n",(bip-aip)/aa);

  /* adjoint test - state deriv, v */
  hist_to_duhgrid(&hist,&mdl,traces.nt/2);

  dynamic_field_rand(&dmdl,DMPVAR);
  // save for later!
  dynamic_field_copy(&dmout,&dmdl);
  static_field_zero(&dmdl);
  
  dynamic_field_rand(&dmin,DMPVAR);
  static_field_zero(&dmin);

  ftsm_v(&dmdl,&mdl,traces.dt);
  aip=dynamic_field_ip(&dmin,&dmdl);
  printf("\nV ONLY:\n");
  printf("<u1,H[c]u0>  =%20.14e\n",aip);
  atsm_v(&dmin,&mdl,traces.dt);
  bip=dynamic_field_ip(&dmin,&dmout);
  printf("<H[c]^Tu1,u0>=%20.14e\n",bip);
  aa=sqrt(dynamic_field_ip(&dmin,&dmin)*dynamic_field_ip(&dmout,&dmout));
  printf("relerr =%20.14e\n",(bip-aip)/aa);

  
/* adjoint test - control deriv */
  hist_to_duhgrid(&hist,&mdl,traces.nt/2);

  static_field_rand(&dmdl,DMPVAR);
  // save a copy!
  static_field_copy(&dmout,&dmdl);
  dynamic_field_zero(&dmdl);

  dynamic_field_rand(&dmin,DMPVAR);
  static_field_zero(&dmin);

  ftsm_p(&dmdl,&mdl,traces.dt);
  aip=dynamic_field_ip(&dmdl,&dmin);
  atsm_p(&dmin,&mdl,traces.dt);
  printf("\nMP ONLY\n");
  printf("<D_cH[c]dc,du>   =%20.14e\n",aip);
  bip=static_field_ip(&dmin,&dmout);
  printf("<dc,D_cH[c]^Tdu> =%20.14e\n",aip,bip); 
  aa=sqrt(static_field_ip(&dmin,&dmin)*static_field_ip(&dmout,&dmout));
  printf("relerr =%20.14e\n",(bip-aip)/aa);
  
/* adjoint test - control deriv, v */
  hist_to_duhgrid(&hist,&mdl,traces.nt/2);

  static_field_rand(&dmdl,DMPVAR);
  // save a copy!
  static_field_copy(&dmout,&dmdl);
  dynamic_field_zero(&dmdl);

  dynamic_field_rand(&dmin,DMPVAR);
  static_field_zero(&dmin);

  ftsm_v(&dmdl,&mdl,traces.dt);
  aip=dynamic_field_ip(&dmdl,&dmin);
  atsm_v(&dmin,&mdl,traces.dt);
  printf("\nMV ONLY\n");
  printf("<D_cH[c]dc,du>   =%20.14e\n",aip);
  bip=static_field_ip(&dmin,&dmout);
  printf("<dc,D_cH[c]^Tdu> =%20.14e\n",aip,bip); 
  aa=sqrt(static_field_ip(&dmin,&dmin)*static_field_ip(&dmout,&dmout));
  printf("relerr =%20.14e\n",(bip-aip)/aa);
  
  // sampler test 
  duhrecs_rand(&dtraces1,DMPVAR);
  duhrecs_zero(&dtraces);
  dynamic_field_rand(&dmdl,DMPVAR);
  dynamic_field_zero(&dmout);
  duhgrid_to_duhrecs(&dtraces,&dmdl,0);
  duhrecs_to_duhgrid(&dtraces1,&dmout,0);

  aip=duhrecs_ip(&dtraces,&dtraces1);
  bip=dynamic_field_ip(&dmdl,&dmout);
  aa=sqrt(duhrecs_ip(&dtraces,&dtraces)*duhrecs_ip(&dtraces1,&dtraces1));

  printf("\nTRACE SAMPLER ONLY\n");
  printf("<t, fwd sample g> = %20.14e\n",aip);
  printf("<adj sample t, g> = %20.14e\n",bip);
  printf("relerr = %20.14e\n",(bip-aip)/aa);

  // adjoint loop test 

  // input:  dmdl=(du,dc), du=0, dc=rand 
  // output: dtraces = 0

  dynamic_field_zero(&dmdl);
  static_field_rand(&dmdl,DMPVAR);
  // no need to save a copy! static part of dmdl remains const

  duhrecs_zero(&dtraces);

  duhgrid_to_duhrecs(&dtraces,&dmdl,0);
  for (it=0;it<traces.nt-1;it++) {
    ftsm_p(&dmdl,&mdl,traces.dt);
    hist_to_duhgrid(&hist,&mdl,it);
    ftsm_v(&dmdl,&mdl,traces.dt);
    // now du at it+1
    duhgrid_to_duhrecs(&dtraces,&dmdl,it+1);
  }
  
  //duhrecs_write(&dtraces,"lin");
  // input:  dtraces1 = rand
  // output: dmin = (du,dc), both zero

  static_field_zero(&dmin);
  dynamic_field_zero(&dmin);
  duhrecs_rand(&dtraces1,DMPVAR);

  // now lu at nt-1 = max it = N in paper
  // update for nt-1=N
  for (it=traces.nt-2;it>-1;it--) {
    // lu at it+1
    hist_to_duhgrid(&hist,&mdl,it);
    duhrecs_to_duhgrid(&dtraces1,&dmin,it+1);
    atsm_v(&dmin,&mdl,traces.dt);
    atsm_p(&dmin,&mdl,traces.dt);
    // now lu at it - update 
  }

  aip=duhrecs_ip(&dtraces,&dtraces1);
  bip=static_field_ip(&dmin,&dmdl);
  printf("\nCOMPLETE LIN & ADJ LIN SIMULATIONS\n");
  printf("<d, DF[c] dc>  = %20.14e\n",aip);
  printf("<DF[c]^Td, dc> = %20.14e\n",bip);
  aa=sqrt(duhrecs_ip(&dtraces,&dtraces)*duhrecs_ip(&dtraces1,&dtraces1));
  printf("relerr = %20.14e\n",(bip-aip)/aa);

  duhgrid_destroy(&mdl);
  duhgrid_destroy(&dmdl);
  duhgrid_destroy(&dmin);
  duhgrid_destroy(&dmout);
  
  duhrecs_destroy(&traces);
  duhrecs_destroy(&dtraces);
}
  
